# coding=utf-8

import os
from os.path import exists, join, abspath, isdir, basename, dirname
from shutil import ignore_patterns

from jinja2 import Template

from guniflask.errors import AbortedError
from guniflask.utils.template import string_lowercase_underscore, string_lowercase_hyphen
from guniflask import __version__


class Command:
    def __init__(self):
        self.exitcode = 0

    @property
    def name(self):
        return ""

    @property
    def syntax(self):
        return ""

    @property
    def short_desc(self):
        return ""

    @property
    def long_desc(self):
        return self.short_desc

    def add_arguments(self, parser):
        pass

    def process_arguments(self, args):
        pass

    def run(self, args):
        raise NotImplementedError


class Step:
    default = None
    desc = None

    def __init__(self):
        self.value = self.default
        self.title = self.desc
        self.remark = self.default

    def process_arguments(self, args):
        pass

    def process_user_input(self, user_input):
        pass


class BaseNameStep(Step):
    desc = 'What is the base name of your application?'

    def process_arguments(self, args):
        project_dir = abspath(args.root_dir or '')
        self.remark = string_lowercase_underscore(basename(project_dir))

    def process_user_input(self, user_input):
        project_basename = string_lowercase_underscore(user_input)
        if not project_basename:
            project_basename = self.remark
        if project_basename and not str.isdigit(project_basename[0]):
            self.value = project_basename
            return {'project_name': project_basename}


class PortStep(Step):
    default = 8000
    desc = 'Would you like to run your application on which port?'

    def process_user_input(self, user_input):
        user_input = user_input.strip()
        if user_input:
            port = self.parse_port(user_input)
        else:
            port = self.parse_port(self.remark)
        if port is not None:
            self.value = port
            return {'port': port}

    @staticmethod
    def parse_port(port):
        try:
            res = int(port)
        except (ValueError, TypeError):
            pass
        else:
            if 0 <= res < 65535:
                return res


class InitCommand(Command):
    @property
    def syntax(self):
        return '[options]'

    @property
    def name(self):
        return 'init'

    @property
    def short_desc(self):
        return 'Initialize a project'

    def add_arguments(self, parser):
        parser.add_argument('-d', '--root-dir', dest='root_dir', metavar='DIR',
                            help='application root directory')

    def run(self, args):
        steps = [BaseNameStep(), PortStep()]
        cur_step, total_steps = 1, 2
        project_dir = abspath(args.root_dir or '')
        self.print_welcome(project_dir)
        print(flush=True)
        settings = {'project_dir': project_dir}
        try:

            for step in steps:
                step.process_arguments(args)
                while True:
                    self.print_request('({}/{}) {}'.format(cur_step, total_steps, step.title),
                                       step.remark)
                    user_input = input()
                    res = step.process_user_input(user_input)
                    if res is None:
                        self.print_go_back_line()
                        continue
                    settings.update(res)
                    self.print_go_back_line()
                    self.print_decision('({}/{}) {}'.format(cur_step, total_steps, step.title),
                                        step.value)
                    cur_step += 1
                    break
        except KeyboardInterrupt:
            print(flush=True)
            return
        try:
            self.copy_files(settings)
        except (KeyboardInterrupt, AbortedError):
            print(flush=True)
            self.print_aborted_error('Process is aborted by user.')
            return
        print(flush=True)
        self.print_success()

    def copy_files(self, settings):
        templates_dir = join(dirname(__file__), 'templates', 'project')
        project_dir = settings['project_dir']
        project_name = settings['project_name']
        settings['project__name'] = string_lowercase_hyphen(project_name)

        print(flush=True)
        print('Copying files:', flush=True)
        self.force = False
        self.copytree(templates_dir, project_dir, settings)

    def copytree(self, src, dst, settings):
        names = os.listdir(src)
        ignored_names = ignore_patterns('*.pyc')(src, names)
        for name in names:
            if name in ignored_names:
                continue
            src_path = join(src, name)
            dst_name = self.render_string(name, **settings)
            dst_path = join(dst, dst_name)
            dst_rel_path = self.relative_path(dst_path, settings['project_dir'])

            if isdir(src_path):
                self.copytree(src_path, dst_path, settings)
            else:
                content = self.read_file(src_path)
                if dst_name.endswith('.tmpl'):
                    dst_name = dst_name[:-5]
                    dst_path = join(dst, dst_name)
                    dst_rel_path = self.relative_path(dst_path, settings['project_dir'])
                    content = self.render_string(content, **settings)
                if exists(dst_path):
                    raw = self.read_file(dst_path)
                    if content == raw:
                        self.print_copying_file('identical', dst_rel_path)
                    else:
                        if self.force:
                            self.print_copying_file('force', dst_rel_path)
                            self.write_file(dst_path, content)
                        else:
                            self.print_copying_file('conflict', dst_rel_path)
                            while True:
                                self.print_request('Overwrite {}?'.format(dst_rel_path), 'Y/n/a/x')
                                user_input = input()
                                user_input = user_input.strip().lower()
                                if not user_input:
                                    user_input = 'y'
                                if user_input[0] not in 'ynax':
                                    self.print_invalid_command()
                                    self.print_go_back_line()
                                    continue
                                user_input = user_input[0]
                                self.print_clean_line()
                                self.print_go_back_line()
                                self.print_decision('Overwrite {}?'.format(dst_rel_path), user_input)
                                if user_input == 'y' or user_input == 'a':
                                    self.print_copying_file('force', dst_rel_path)
                                    self.write_file(dst_path, content)
                                    if user_input == 'a':
                                        self.force = True
                                elif user_input == 'n':
                                    self.print_copying_file('skip', dst_rel_path)
                                elif user_input == 'x':
                                    raise AbortedError
                                break
                else:
                    self.print_copying_file('create', dst_rel_path)
                    self.write_file(dst_path, content)

    @staticmethod
    def relative_path(path, fpath):
        path = abspath(path)
        fpath = abspath(fpath)
        if path.startswith(fpath):
            path = path[len(fpath):]
            if path.startswith(os.path.sep):
                path = path[1:]
        return path

    @staticmethod
    def read_file(path):
        with open(path, 'r', encoding='utf-8') as f:
            return f.read()

    @staticmethod
    def write_file(path, raw):
        d = dirname(path)
        if not exists(d):
            os.makedirs(d)
        with open(path, 'w', encoding='utf-8') as f:
            f.write(raw)

    @staticmethod
    def render_string(raw, **kwargs):
        return Template(raw, keep_trailing_newline=True).render(**kwargs)

    @staticmethod
    def print_request(question, option):
        print('\033[32;40m?\033[0m {} \033[37;40m({})\033[0m '
              .format(question, option), end='', flush=True)

    @staticmethod
    def print_decision(question, decision):
        print('\033[32;40m?\033[0m {} \033[36;40m{}\033[0m'
              .format(question, decision), flush=True)

    @staticmethod
    def print_welcome(project_dir):
        print('\033[37;40mWelcome to gunicorn generator\033[0m \033[33;40mv{}\033[0m'
              .format(__version__), flush=True)
        print('\033[37;40mApplication file will be created in folder:\033[0m \033[33;40m{}\033[0m'
              .format(project_dir), flush=True)

    @staticmethod
    def print_go_back_line():
        print('\033[1A\r\033[K', end='', flush=True)

    @staticmethod
    def print_clean_line():
        print('\r\033[K', end='', flush=True)

    @staticmethod
    def print_invalid_command():
        print('\033[31;40m>>\033[0m Please enter a valid command', end='', flush=True)

    @staticmethod
    def print_success():
        print('\033[32;40mApplication is created successfully.\033[0m', flush=True)

    @staticmethod
    def print_aborted_error(message):
        print('\033[33;40m{}\033[0m'.format(message), flush=True)

    @staticmethod
    def print_copying_file(t, path):
        color = 0
        if t == 'identical':
            color = 36
        elif t == 'conflict':
            color = 31
        elif t == 'create':
            color = 32
        elif t == 'force' or t == 'skip':
            color = 33
        print('\033[{};40m{:>9}\033[0m {}'.format(color, t, path), flush=True)


class VersionCommand(Command):
    @property
    def name(self):
        return "version"

    @property
    def short_desc(self):
        return "Print the version"

    def run(self, args):
        print("guniflask version {}".format(__version__))



# coding=utf-8

from guniflask.utils.template import *


def test_string_camelcase():
    assert string_camelcase('ABc-dE FgH') == 'AbcDeFgh'


def test_string_lowercase_hyphen():
    assert string_lowercase_hyphen(' AB_cD-Ef gh ') == 'ab-cd-ef-gh'


def test_string_lowercase_underscore():
    assert string_lowercase_underscore(' AB_cD-Ef gh ') == 'ab_cd_ef_gh'


def test_string_uppercase_underscore():
    assert string_uppercase_underscore(' AB_cD-Ef gh ') == 'AB_CD_EF_GH'



# coding=utf-8

from flask import Blueprint

hello_world_blueprint = Blueprint('hello_world', __name__, url_prefix='/api')


@hello_world_blueprint.route('/hello-world', methods=['GET'])
def hello_world():
    return 'Hello World!'



# coding=utf-8


def make_settings(app, settings):
    """
    This function is invoked before initializing app.
    """


def init_app(app, settings):
    """
    This function is invoked before running app.
    """



# coding=utf-8

import os
from os.path import isfile, join, isdir, abspath
from importlib import import_module
from pkgutil import iter_modules
import inspect


def load_config(fname):
    if fname is None or not isfile(fname):
        raise FileNotFoundError("Cannot find configuration file '{}'".format(fname))
    code = compile(open(fname, 'rb').read(), fname, 'exec')
    cfg = {
        "__builtins__": __builtins__,
        "__name__": "__config__",
        "__file__": fname,
        "__doc__": None,
        "__package__": None
    }
    exec(code, cfg, cfg)
    return cfg


def load_profile_config(conf_dir, name, profiles=None):
    pc = {}
    if profiles:
        profiles = profiles.split(',')
        profiles.reverse()
        for profile in profiles:
            if profile:
                pc_file = join(conf_dir, name + '-' + profile + '.py')
                if not isfile(pc_file):
                    pc_file = join(conf_dir, name + '_' + profile + '.py')
                if isfile(pc_file):
                    c = load_config(pc_file)
                    pc.update(c)
    return pc


app_default_settings = {
    'active_profiles': None,
    'debug': False,
    'cors': True,
    # Flask-SQLAlchemy
    'SQLALCHEMY_TRACK_MODIFICATIONS': False
}


def load_app_config(name):
    from guniflask.settings import Settings

    c = app_default_settings
    conf_dir = os.environ.get('GUNIFLASK_CONF_DIR')
    if conf_dir:
        c.update(load_config(join(conf_dir, name + '.py')))
        active_profiles = os.environ.get('GUNIFLASK_ACTIVE_PROFILES', c.get('active_profiles'))
        c.update(load_profile_config(conf_dir, name, profiles=active_profiles))
        c['active_profiles'] = active_profiles
    settings = Settings()
    for name in c:
        if not name.startswith('_') and not inspect.ismodule(c[name]) and not inspect.isfunction(c[name]):
            settings[name] = c[name]
    if os.environ.get('GUNIFLASK_DEBUG'):
        settings['debug'] = True
    return settings


def walk_modules(path):
    mods = []
    mod = import_module(path)
    mods.append(mod)
    if hasattr(mod, '__path__'):
        for _, subpath, ispkg in iter_modules(mod.__path__):
            fullpath = path + '.' + subpath
            if ispkg:
                mods += walk_modules(fullpath)
            else:
                submod = import_module(fullpath)
                mods.append(submod)
    return mods


def walk_files(path):
    path = abspath(path)
    files = []
    if isdir(path):
        names = os.listdir(path)
        for name in names:
            files += walk_files(join(path, name))
    elif isfile(path):
        files.append(path)
    return files



# coding=utf-8

import re

_camelcase_invalid_chars = re.compile(r'[^a-zA-Z\d]')

_name_invalid_chars = re.compile(r'[^a-zA-Z\d\-_]')


def string_camelcase(s):
    return _camelcase_invalid_chars.sub('', s.title())


def string_lowercase_hyphen(s):
    return _name_invalid_chars.sub('', s.strip().replace(' ', '-').lower()).replace('_', '-')


def string_lowercase_underscore(s):
    return _name_invalid_chars.sub('', s.strip().replace(' ', '_').lower()).replace('-', '_')


def string_uppercase_underscore(s):
    return _name_invalid_chars.sub('', s.strip().replace(' ', '_').upper()).replace('-', '_')



# coding=utf-8

from os.path import join
from collections import defaultdict
from keyword import iskeyword

from sqlalchemy import ForeignKeyConstraint, UniqueConstraint, PrimaryKeyConstraint, CheckConstraint, ForeignKey
import inflect

from guniflask.utils.template import string_camelcase, string_lowercase_underscore
from guniflask.dialects import supported_dialects

inflect_engine = inflect.engine()


class SqlToModelGenerator:
    def __init__(self, name, metadata, indent=4):
        dialect_name = metadata.bind.dialect.name
        if dialect_name not in supported_dialects:
            raise ValueError('Unsupported dialect: {}'.format(dialect_name))

        self.dialect = supported_dialects[dialect_name]
        self.name = name
        self.indent = ' ' * indent

        many_to_many_tables = set()
        many_to_many_links = defaultdict(list)
        for table in metadata.tables.values():
            fk_constraints = [i for i in table.constraints if isinstance(i, ForeignKeyConstraint)]
            if len(fk_constraints) == 2 and all(col.foreign_keys for col in table.columns):
                many_to_many_tables.add(table.name)
                tablename = sorted(fk_constraints, key=get_constraint_sort_key)[0].elements[0].column.table.name
                many_to_many_links[tablename].append(table)

        self.models = {}
        for table in metadata.sorted_tables:
            if table.name in many_to_many_tables:
                continue
            self.models[table.name] = Model(table, many_to_many_links[table.name])

            # Add many-to-one relations
            for constraint in sorted(table.constraints, key=get_constraint_sort_key):
                if isinstance(constraint, ForeignKeyConstraint):
                    tablename = constraint.elements[0].column.table.name
                    self.models[tablename].add_many_to_one_relation(constraint)

    def render(self, path):
        with open(join(path, '__init__.py'), 'w', encoding='utf-8') as f:
            f.write('# coding=utf-8\n')
        for model in self.models.values():
            model_name = convert_to_valid_identifier(model.table.name)
            with open(join(path, model_name + '.py'), 'w', encoding='utf-8') as f:
                f.write('# coding=utf-8\n\n')
                f.write(self.render_imports())
                tables_content = self.render_secondary_tables(model)
                if tables_content:
                    f.write('\n')
                    f.write(tables_content)
                f.write('\n\n')
                f.write(self.render_model(model))

    def render_imports(self):
        return 'from {} import db\n'.format(self.name)

    def render_model(self, model):
        header_str = "class {}(db.Model):\n".format(model.class_name)
        header_str += "{}__tablename__ = '{}'\n\n".format(self.indent, model.table.name)
        columns_str = ''
        for col in model.table.columns:
            attr = convert_to_valid_identifier(col.name)
            show_name = attr != col.name
            columns_str += '{}{} = {}\n'.format(self.indent, attr,
                                                self.render_column(col, show_name=show_name))
        relationships_str = ''
        for r in model.relationships:
            relationships_str += self.indent + self.render_relationship(r) + '\n'
        return header_str + columns_str + relationships_str

    def render_secondary_tables(self, model):
        return '\n'.join([self.render_table(r.association_table) for r in model.relationships
                          if isinstance(r, ManyToManyRelationship)])

    def render_table(self, table):
        columns_str = ',\n'.join(self.indent + self.render_column(col, show_name=True) for col in table.columns)
        tablename = convert_to_valid_identifier(table.name)
        return '{} = db.Table({!r},\n{}\n)\n'.format(tablename, table.name, columns_str)

    def render_column(self, column, show_name=False):
        is_sole_pk = column.primary_key and len(column.table.primary_key) == 1
        dedicated_fks = [c for c in column.foreign_keys if len(c.constraint.columns) == 1]
        is_unique = any(isinstance(c, UniqueConstraint) and set(c.columns) == {column}
                        for c in column.table.constraints)
        is_unique = is_unique or any(i.unique and set(i.columns) == {column}
                                     for i in column.table.indexes)
        has_index = any(set(i.columns) == {column} for i in column.table.indexes)

        kwargs = []
        if column.key != column.name:
            kwargs.append('key')
        if column.primary_key:
            kwargs.append('primary_key')
        if column.autoincrement is True:
            kwargs.append('autoincrement')
        if not column.nullable and not is_sole_pk:
            kwargs.append('nullable')
        if is_unique:
            column.unique = True
            kwargs.append('unique')
        if has_index:
            column.index = True
            kwargs.append('index')
        if column.comment:
            kwargs.append('comment')

        kwargs_str = ''
        for k in kwargs:
            kwargs_str += ', {}={!r}'.format(k, getattr(column, k))
        return "db.Column({})".format(', '.join(([repr(column.name)] if show_name else []) +
                                                [self.render_column_type(column.type)] +
                                                [self.render_constraint(x) for x in dedicated_fks] +
                                                ['{}={!r}'.format(i, getattr(column, i)) for i in kwargs]))

    def render_column_type(self, coltype):
        return 'db.{}'.format(self.dialect.convert_column_type(coltype))

    def render_constraint(self, constraint):
        def render_fk_options(*args):
            opts = [repr(i) for i in args]
            for attr in 'ondelete', 'onupdate':
                value = getattr(constraint, attr, None)
                if value:
                    opts.append('{}={!r}'.format(attr, value))
            return ', '.join(opts)

        if isinstance(constraint, ForeignKey):
            remote_column = '{}.{}'.format(constraint.column.table.name,
                                           constraint.column.name)
            return 'db.ForeignKey({})'.format(render_fk_options(remote_column))

    def render_relationship(self, relationship):
        kwargs_str = ', '.join([repr(relationship.target_cls)] +
                               ['{}={}'.format(i, relationship.kwargs[i]) for i in
                                sorted(relationship.kwargs.keys())])
        return '{} = db.relationship({})'.format(relationship.preferred_name, kwargs_str)


def convert_to_valid_identifier(name):
    name = string_lowercase_underscore(name)
    if name[0].isdigit() or iskeyword(name):
        name = '_' + name
    return name


def get_constraint_sort_key(constraint):
    if isinstance(constraint, CheckConstraint):
        return 'C{0}'.format(constraint.sqltext)
    return constraint.__class__.__name__[0] + repr(list(constraint.columns.keys()))


class Model:
    parent_name = 'db.Model'

    def __init__(self, table, association_tables):
        self.table = table
        self.class_name = self.table_name_to_class_name(table.name)
        self.relationships = []

        # Add many-to-many relationships
        for association_table in association_tables:
            fk_constraints = [c for c in association_table.constraints if isinstance(c, ForeignKeyConstraint)]
            fk_constraints.sort(key=get_constraint_sort_key)
            target_cls = self.table_name_to_class_name(fk_constraints[1].elements[0].column.table.name)
            relationship = ManyToManyRelationship(self.class_name, target_cls, association_table)
            self.relationships.append(relationship)

    @staticmethod
    def table_name_to_class_name(table_name):
        name = string_camelcase(table_name)
        if name[0].isdigit():
            name = '_' + name
        return name

    def add_many_to_one_relation(self, constraint):
        target_cls = self.table_name_to_class_name(constraint.table.name)
        relationship = ManyToOneRelationship(self.class_name, target_cls, constraint)
        self.relationships.append(relationship)


class Relationship:
    def __init__(self, source_cls, target_cls):
        self.source_cls = source_cls
        self.target_cls = target_cls
        self.kwargs = {}
        self.preferred_name = None


class ManyToOneRelationship(Relationship):
    def __init__(self, source_cls, target_cls, constraint):
        super().__init__(source_cls, target_cls)

        self.preferred_name = convert_to_valid_identifier(target_cls)
        self.constraint = constraint

        # Add uselist=False to one-to-one relationships
        column_names = list(constraint.columns.keys())
        if any(isinstance(c, (PrimaryKeyConstraint, UniqueConstraint)) and
               set(col.name for col in c.columns) == set(column_names)
               for c in constraint.table.constraints):
            self.kwargs['uselist'] = 'False'

        if self.kwargs.get('uselist') is not False:
            self.preferred_name = inflect_engine.plural(self.preferred_name)
            self.kwargs['lazy'] = repr('select')
        else:
            self.kwargs['lazy'] = repr('joined')
        self.kwargs['backref'] = "db.backref({}, lazy='joined')".format(
            repr(convert_to_valid_identifier(source_cls)))


class ManyToManyRelationship(Relationship):
    def __init__(self, source_cls, target_cls, association_table):
        super().__init__(source_cls, target_cls)

        self.preferred_name = inflect_engine.plural(convert_to_valid_identifier(target_cls))
        self.association_table = association_table

        self.kwargs['secondary'] = convert_to_valid_identifier(association_table.name)
        self.kwargs['lazy'] = repr('select')
        self.kwargs['backref'] = "db.backref({}, lazy='select')".format(
            repr(inflect_engine.plural(convert_to_valid_identifier(source_cls))))



# coding=utf-8

import sqlalchemy

__all__ = ['model_to_dict', 'dict_to_model']


def model_to_dict(model):
    col_attrs = sqlalchemy.inspect(model).mapper.column_attrs
    return {c.key: getattr(model, c.key) for c in col_attrs}


def dict_to_model(dict_obj, model_cls=None):
    if model_cls is None:
        return dict(dict_obj)
    col_attrs = sqlalchemy.inspect(model_cls).mapper.column_attrs
    kwargs = {c.key: dict_obj[c.key] for c in col_attrs if c.key in dict_obj}
    model = model_cls(**kwargs)
    return model
